/*
 Soekris Net4801 GPIO driver
 Provides basic control (read and write) over the 12 general purpose
 IO-pins of the Soekris Net4801 and the error-led (which happens to
 be connected to another general purpose IO pin)
 Also provides read access to the termperature and voltage sensors

  (c) Copyright 2003-2004   Martin Hejl <martin@hejl.de>
                            G&H Softwareentwicklung GmbH

 **********************************************************************
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 **********************************************************************
 *
 */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
	#include <linux/modversions.h>
#endif

#include "4801driver.h"
static void _net4801_cleanup(void);
int net4801_init(struct gpio_operations **);

extern unsigned long  fromString(char* , int );
extern int toString(unsigned long , char* , int );

/* From http://phk.freebsd.dk/soekris/env4801/env4801.c 
According to the author the values in the following two arrays were
"guessed and measured in my lab."
But the data they produce looks sensible, so we'll use this until 
there's official documentation from Soekris Engineering
*/
static double sc[14] = {
	1.0, 1.0, 2.0, 20.0,
	1.0, 1.0, 1.0, 2.0,
	2.0, 1.0, 2.0, 1.0,
	1.0, 1.0
};

static char const *nm[14] = {
	"avi0",
	"avi1",
	"VCC",
	"VPWR",
	"avi4",
	"avi5",
	"avi6",
	"Vsb",
	"Vdd",
	"Vbat",
	"AVdd",
	"ts0",
	"ts1",
	"ts2"
};

/* 
variable which holds the currently selected device on the PC87366 
(GPIO, TMS, VLM) so we don't keep selecting an already selected device
*/
static unsigned int _selected_device = 0;

static void device_select(unsigned int dev) {
	if (dev==_selected_device) return;

	outb(SIO_DEV_SEL, SIO_INDEX);
	outb(dev, SIO_DATA);
	_selected_device = dev;
}

static void _net4801_readTemperature(
	int index,
	unsigned char* status,
	unsigned char* value)
{
	device_select(SIO_VLM_DEV);

	/* bank select */
	outb(index, sio_tms_base + SIO_TMSBS);

	/* Read the status */
	*status = inb(sio_tms_base + SIO_TCHCFST);

	/* Read the temperature */
	*value = inb(sio_tms_base + SIO_RDCHT);
}

static void _net4801_readVoltage(
	int index,
	unsigned char* value)
{

	device_select(SIO_VLM_DEV);

	/* bank select */
	outb(index, sio_vlm_base + SIO_VLMBS);

	/* Read the voltage */
	*value  = inb(sio_vlm_base + SIO_RDCHV);


}

static int _net4801_readTemperature_string(char* buffer,int len) {
	static int i;
	static unsigned char status;
	static unsigned char value;
	static int offs;

	offs = 0;
	status = 0;
	value = 0;

	for(i = 0; i < SIO_TMS_BANKS; i++) {
		_net4801_readTemperature(i,&status, &value);
		offs += sprintf(buffer+offs, "Temp %d 0x%x %d C\n", i, status, value);

		/* buffer overflow - should never happen!! */
		if (offs>=len-1)
			break;
	}
	buffer[offs]='\0';
	return offs+1;
}

static int _net4801_readVoltage_string(char* buffer,int len) {
	static int i;
	static unsigned char value;
	static int offs;
	static double a;
	static int d1;
	static int d2;
	static int d3;

	offs = 0;
	for(i = 0; i < SIO_LVM_BANKS; i++) {
		_net4801_readVoltage(i,&value);

		/* see 11.2.1 of PC87366 Spec, (page 177) */
		a = value*SIO_LVM_VREF*2.45/256.0;

		/* convert it into two decimals */
		d1 = a*PRECISION*sc[i];
		d2 = d1/PRECISION;
		d3 = d1 - PRECISION*d2;

		offs += sprintf(buffer+offs, "%s %d.%s%d V\n", nm[i], d2,d3<10?"0":"",d3);

		/* buffer overflow - should never happen!! */
		if (offs>=len)
			break;
	}

	return offs;
}

struct gpio_operations driver4801_ops = {
	.init                = net4801_init,
	.cleanup             = _net4801_cleanup,
	.readTemperature     = _net4801_readTemperature_string,
	.readVoltage         = _net4801_readVoltage_string
};

int net4801_init(struct gpio_operations **driver_ops)
{
	static struct pci_dev *bridge;
	static unsigned base;
	static int i;

	if ((bridge = pci_find_device(SC1100_VENDOR_ID,
					SC1100_F0_DEVICE_ID ,
					NULL)) == NULL)
		return -ENODEV;

	base = 0x6100;

	if (request_region(base, GEODE_GPIO_SIZE, "Soekris net5501 env") == 0) {
		printk(KERN_ERR OUR_NAME ": can't allocate I/O for Geode GPIOs\n");
		return -EBUSY;
	}

	printk(OUR_NAME ": Soekris net5501 env driver Version " VERSION " (C) 2003 Martin Hejl\n");

	geode_gpio_base = base;

	offset_87336 = 0;
	outb(SIO_SID,SIO_INDEX);
	if (inb(SIO_DATA) != SIO_SID_VALUE)
	{
		offset_87336 = SIO_BASE_OFFSET;
		outb(SIO_SID,SIO_INDEX);
		if (inb(SIO_DATA) != SIO_SID_VALUE)
		{
			printk(KERN_ERR OUR_NAME ": PC87336 not detected\n" );

			/* clean up */
			//unregister_chrdev(gpio_major, OUR_NAME);
			release_region(geode_gpio_base, GEODE_GPIO_SIZE);
			return -EBUSY;
		}
	}

	// select temperature sensors
	device_select(SIO_TMS_DEV);

	// enable
	sio_tms_base = 0;
	outb(SIO_DEV_ENB,SIO_INDEX);
	if (inb(SIO_DATA)==0) {
		printk(OUR_NAME ": TMS disabled\n");
		driver4801_ops.readTemperature = NULL;
	} else {
		// read TMS base address
		outb(SIO_BASE_HADDR,SIO_INDEX );
		sio_tms_base = inb(SIO_DATA) << 8;

		outb(SIO_BASE_LADDR,SIO_INDEX);
		sio_tms_base |= inb(SIO_DATA);
	}

	// select voltage monitors
	device_select(SIO_VLM_DEV);

	// enable
	sio_vlm_base = 0;
	outb(SIO_DEV_ENB,SIO_INDEX);
	if (inb(SIO_DATA)==0) {
		printk(OUR_NAME ": VLM disabled\n");
		driver4801_ops.readVoltage = NULL;
	} else {
		outb(SIO_BASE_HADDR, SIO_INDEX);
		sio_vlm_base = inb(SIO_DATA) << 8;

		outb(SIO_BASE_LADDR, SIO_INDEX);
		sio_vlm_base  |= inb(SIO_DATA);
	}

	// obtain the specified VLM/TMS accuracy (PC87366 Spec, page 208)
	if (sio_tms_base!=0) {
		outb(0x00, sio_tms_base + 0x8 );
		outb(0x0f, sio_tms_base + 0x9 );
		outb(0x08, sio_tms_base + 0xa );
		outb(0x04, sio_tms_base + 0xb );
		outb(0x35, sio_tms_base + 0xc );
		outb(0x05, sio_tms_base + 0xd );
		outb(0x05, sio_tms_base + 0xe );
	}

	/* configure to
	 * Internal Vref,
	 * Monitoring operation enabled
	 */
	outb(0, sio_tms_base + SIO_TMSCFG);

	/* for all  banks */
	for(i = 0; i < SIO_TMS_BANKS; i++) {
		/* select bank */
		outb(i,sio_tms_base + SIO_TMSBS);

		/* enable channel */
		outb(1, sio_tms_base + SIO_TCHCFST);
	}

	/* set sampling delay to 160 us */
	outb(0x10, sio_vlm_base + SIO_VCNVR);

	/* configure to
	 * Internal Vref,
	 * Monitoring operation enabled
	 */
	outb(0, sio_vlm_base + SIO_VLMCFG);

	/* for all banks */
	for(i = 0; i < SIO_LVM_BANKS; i++) {

		/* select bank */
		outb(i, sio_vlm_base + SIO_VLMBS);

		/* enable channel */
		outb(1, sio_vlm_base + SIO_VCHCFST);
	}

	*driver_ops = &driver4801_ops;
	return 0;
}

static void _net4801_cleanup(void)
{
	release_region(geode_gpio_base, GEODE_GPIO_SIZE);
}
